@@ -9,7 +9,6 @@ use core::mem::ManuallyDrop;
 
 use std::error::Error as StdError;
 
-use atty::Stream;
 use once_cell::sync::OnceCell;
 
 #[allow(unreachable_pub)]
@@ -27,7 +26,12 @@ pub use ReportHandler as EyreContext;
 #[allow(unreachable_pub)]
 pub use WrapErr as Context;
 
-use crate::{Diagnostic, GraphicalReportHandler, NarratableReportHandler};
+#[cfg(not(feature = "fancy"))]
+use crate::DebugReportHandler;
+use crate::Diagnostic;
+#[cfg(feature = "fancy")]
+use crate::MietteHandler;
+
 use error::ErrorImpl;
 
 mod context;
@@ -88,19 +92,10 @@ fn capture_handler(error: &(dyn Diagnostic + 'static)) -> Box<dyn ReportHandler>
 }
 
 fn get_default_printer(_err: &(dyn Diagnostic + 'static)) -> Box<dyn ReportHandler + 'static> {
-    let fancy = if let Ok(string) = std::env::var("NO_COLOR") {
-        string == "0"
-    } else if let Ok(string) = std::env::var("CLICOLOR") {
-        string != "0" || string == "1"
-    } else {
-        atty::is(Stream::Stdout) && atty::is(Stream::Stderr) && !ci_info::is_ci()
-    };
-    let size = term_size::dimensions().unwrap_or((80, 0)).0;
-    if fancy {
-        Box::new(GraphicalReportHandler::new().with_width(size))
-    } else {
-        Box::new(NarratableReportHandler)
-    }
+    #[cfg(feature = "fancy")]
+    return Box::new(MietteHandler::new());
+    #[cfg(not(feature = "fancy"))]
+    return Box::new(DebugReportHandler::new());
 }
 
 impl dyn ReportHandler {
@@ -153,7 +148,7 @@ pub trait ReportHandler: core::any::Any + Send + Sync {
     /// impl ReportHandler for Handler {
     ///     fn debug(
     ///         &self,
-    ///         error: &(dyn Diagnostic + 'static),
+    ///         error: &dyn Diagnostic,
     ///         f: &mut core::fmt::Formatter<'_>,
     ///     ) -> core::fmt::Result {
     ///         use core::fmt::Write as _;
@@ -170,7 +165,7 @@ pub trait ReportHandler: core::any::Any + Send + Sync {
     /// ```
     fn debug(
         &self,
-        error: &(dyn Diagnostic + 'static),
+        error: &(dyn Diagnostic),
         f: &mut core::fmt::Formatter<'_>,
     ) -> core::fmt::Result;
 
