@@ -1,12 +1,11 @@
 use std::fmt::{self, Write};
 
 use owo_colors::{OwoColorize, Style};
-use unicode_width::UnicodeWidthStr;
 
 use crate::chain::Chain;
 use crate::handlers::theme::*;
-use crate::protocol::{Diagnostic, DiagnosticSnippet, Severity};
-use crate::{ReportHandler, SourceSpan, SpanContents};
+use crate::protocol::{Diagnostic, Severity};
+use crate::{LabeledSpan, MietteError, ReportHandler, SourceCode, SourceSpan, SpanContents};
 
 /**
 A [ReportHandler] that displays a given [crate::Report] in a quasi-graphical
@@ -24,6 +23,8 @@ pub struct GraphicalReportHandler {
     pub(crate) linkify_code: bool,
     pub(crate) termwidth: usize,
     pub(crate) theme: GraphicalTheme,
+    pub(crate) footer: Option<String>,
+    pub(crate) context_lines: usize,
 }
 
 impl GraphicalReportHandler {
@@ -34,6 +35,8 @@ impl GraphicalReportHandler {
             linkify_code: true,
             termwidth: 200,
             theme: GraphicalTheme::default(),
+            footer: None,
+            context_lines: 1,
         }
     }
 
@@ -43,12 +46,14 @@ impl GraphicalReportHandler {
             linkify_code: true,
             termwidth: 200,
             theme,
+            footer: None,
+            context_lines: 1,
         }
     }
 
-    /// Disables error code linkification using [Diagnostic::url].
-    pub fn without_code_linking(mut self) -> Self {
-        self.linkify_code = false;
+    /// Whether to enable error code linkification using [Diagnostic::url].
+    pub fn with_links(mut self, links: bool) -> Self {
+        self.linkify_code = links;
         self
     }
 
@@ -63,6 +68,18 @@ impl GraphicalReportHandler {
         self.termwidth = width;
         self
     }
+
+    /// Sets the "global" footer for this handler.
+    pub fn with_footer(mut self, footer: String) -> Self {
+        self.footer = Some(footer);
+        self
+    }
+
+    /// Sets the number of lines of context to show around each error.
+    pub fn with_context_lines(mut self, lines: usize) -> Self {
+        self.context_lines = lines;
+        self
+    }
 }
 
 impl Default for GraphicalReportHandler {
@@ -84,67 +101,65 @@ impl GraphicalReportHandler {
         self.render_header(f, diagnostic)?;
         writeln!(f)?;
         self.render_causes(f, diagnostic)?;
-
-        if let Some(snippets) = diagnostic.snippets() {
-            for snippet in snippets {
-                writeln!(f)?;
-                self.render_snippet(f, &snippet)?;
-            }
-        }
-
+        self.render_snippets(f, diagnostic)?;
         self.render_footer(f, diagnostic)?;
+        self.render_related(f, diagnostic)?;
+        if let Some(footer) = &self.footer {
+            writeln!(f)?;
+            let width = self.termwidth.saturating_sub(4);
+            let opts = textwrap::Options::new(width)
+                .initial_indent("  ")
+                .subsequent_indent("  ");
+            writeln!(f, "{}", textwrap::fill(footer, opts))?;
+        }
         Ok(())
     }
 
     fn render_header(&self, f: &mut impl fmt::Write, diagnostic: &(dyn Diagnostic)) -> fmt::Result {
+        let severity_style = match diagnostic.severity() {
+            Some(Severity::Error) | None => self.theme.styles.error,
+            Some(Severity::Warning) => self.theme.styles.warning,
+            Some(Severity::Advice) => self.theme.styles.advice,
+        };
         let mut header = String::new();
-        let mut width = 0;
-        write!(
-            header,
-            "{}",
-            self.theme.characters.hbar.to_string().repeat(4)
-        )?;
-        width += UnicodeWidthStr::width(&header[..]);
-        if self.linkify_code && diagnostic.url().is_some() && diagnostic.code().is_some() {
+        if self.linkify_code && diagnostic.url().is_some() {
             let url = diagnostic.url().unwrap(); // safe
-            let code = format!(
-                "{} (click for details)",
-                diagnostic
-                    .code()
-                    .expect("MIETTE BUG: already got checked for None")
+            let code = if let Some(code) = diagnostic.code() {
+                format!("{} ", code)
+            } else {
+                "".to_string()
+            };
+            let link = format!(
+                "\u{1b}]8;;{}\u{1b}\\{}\u{1b}]8;;\u{1b}\\",
+                url,
+                format!(
+                    "{}{}",
+                    code.style(severity_style),
+                    "(link)".style(self.theme.styles.link)
+                )
             );
-            let link = format!("\u{1b}]8;;{}\u{1b}\\{}\u{1b}]8;;\u{1b}\\", url, code);
-            write!(header, "[{}]", link.style(self.theme.styles.code))?;
-            width += UnicodeWidthStr::width(&url.to_string()[..])
-                + UnicodeWidthStr::width(&code[..])
-                + 2;
+            write!(header, "{}", link)?;
+            writeln!(f, "{}", header)?;
         } else if let Some(code) = diagnostic.code() {
-            write!(header, "[{}]", code.style(self.theme.styles.code))?;
-            width += UnicodeWidthStr::width(&code.to_string()[..]) + 2;
+            write!(header, "{}", code.style(severity_style),)?;
+            if let Some(link) = diagnostic.url() {
+                write!(header, " ({})", link.style(self.theme.styles.link))?;
+            }
+            writeln!(f, "{}", header)?;
         }
-        writeln!(
-            f,
-            "{}{}",
-            header,
-            self.theme.characters.hbar.to_string().repeat(
-                self.termwidth
-                    .saturating_sub(width)
-                    .saturating_sub("Error: ".len())
-            ),
-        )?;
         Ok(())
     }
 
     fn render_causes(&self, f: &mut impl fmt::Write, diagnostic: &(dyn Diagnostic)) -> fmt::Result {
         let (severity_style, severity_icon) = match diagnostic.severity() {
-            Some(Severity::Error) | None => (self.theme.styles.error, self.theme.characters.x),
-            Some(Severity::Warning) => (self.theme.styles.warning, self.theme.characters.warning),
-            Some(Severity::Advice) => (self.theme.styles.advice, self.theme.characters.point_right),
+            Some(Severity::Error) | None => (self.theme.styles.error, &self.theme.characters.error),
+            Some(Severity::Warning) => (self.theme.styles.warning, &self.theme.characters.warning),
+            Some(Severity::Advice) => (self.theme.styles.advice, &self.theme.characters.advice),
         };
 
-        let initial_indent = format!("    {} ", severity_icon.style(severity_style));
-        let rest_indent = format!("    {} ", self.theme.characters.vbar.style(severity_style));
-        let width = self.termwidth.saturating_sub(4);
+        let initial_indent = format!("  {} ", severity_icon.style(severity_style));
+        let rest_indent = format!("  {} ", self.theme.characters.vbar.style(severity_style));
+        let width = self.termwidth.saturating_sub(2);
         let opts = textwrap::Options::new(width)
             .initial_indent(&initial_indent)
             .subsequent_indent(&rest_indent);
@@ -154,20 +169,28 @@ impl GraphicalReportHandler {
         if let Some(cause) = diagnostic.source() {
             let mut cause_iter = Chain::new(cause).peekable();
             while let Some(error) = cause_iter.next() {
-                let char = if cause_iter.peek().is_some() {
+                let is_last = cause_iter.peek().is_none();
+                let char = if !is_last {
                     self.theme.characters.lcross
                 } else {
                     self.theme.characters.lbot
                 };
                 let initial_indent = format!(
-                    "    {}{}{} ",
+                    "  {}{}{} ",
                     char, self.theme.characters.hbar, self.theme.characters.rarrow
                 )
                 .style(severity_style)
                 .to_string();
-                let rest_indent = format!("    {}   ", self.theme.characters.vbar)
-                    .style(severity_style)
-                    .to_string();
+                let rest_indent = format!(
+                    "  {}   ",
+                    if is_last {
+                        ' '
+                    } else {
+                        self.theme.characters.vbar
+                    }
+                )
+                .style(severity_style)
+                .to_string();
                 let opts = textwrap::Options::new(width)
                     .initial_indent(&initial_indent)
                     .subsequent_indent(&rest_indent);
@@ -180,35 +203,120 @@ impl GraphicalReportHandler {
 
     fn render_footer(&self, f: &mut impl fmt::Write, diagnostic: &(dyn Diagnostic)) -> fmt::Result {
         if let Some(help) = diagnostic.help() {
-            let help = help.style(self.theme.styles.help);
-            writeln!(f)?;
             let width = self.termwidth.saturating_sub(4);
-            let initial_indent = format!("    {} ", self.theme.characters.fyi);
+            let initial_indent = "  help: ".style(self.theme.styles.help).to_string();
             let opts = textwrap::Options::new(width)
                 .initial_indent(&initial_indent)
-                .subsequent_indent("      ");
+                .subsequent_indent("        ");
             writeln!(f, "{}", textwrap::fill(&help.to_string(), opts))?;
         }
         Ok(())
     }
 
-    fn render_snippet(
+    fn render_related(
         &self,
         f: &mut impl fmt::Write,
-        snippet: &DiagnosticSnippet<'_>,
+        diagnostic: &(dyn Diagnostic),
     ) -> fmt::Result {
-        let (contents, lines) = self.get_lines(snippet)?;
+        if let Some(related) = diagnostic.related() {
+            writeln!(f)?;
+            for rel in related {
+                write!(f, "Error: ")?;
+                self.render_header(f, rel)?;
+                writeln!(f)?;
+                self.render_causes(f, rel)?;
+                self.render_snippets(f, rel)?;
+                self.render_footer(f, rel)?;
+                self.render_related(f, rel)?;
+            }
+        }
+        Ok(())
+    }
 
-        // Highlights are the bits we're going to underline in our overall
-        // snippet, and we need to do some analysis first to come up with
-        // gutter size.
-        let mut highlights = snippet.highlights.clone().unwrap_or_else(Vec::new);
-        // sorting is your friend.
-        highlights.sort_unstable_by_key(|(_, h)| h.offset());
-        let highlights = highlights
-            .into_iter()
+    fn render_snippets(
+        &self,
+        f: &mut impl fmt::Write,
+        diagnostic: &(dyn Diagnostic),
+    ) -> fmt::Result {
+        if let Some(source) = diagnostic.source_code() {
+            if let Some(labels) = diagnostic.labels() {
+                let mut labels = labels.collect::<Vec<_>>();
+                labels.sort_unstable_by_key(|l| l.inner().offset());
+                if !labels.is_empty() {
+                    let contents = labels
+                        .iter()
+                        .map(|label| {
+                            source.read_span(label.inner(), self.context_lines, self.context_lines)
+                        })
+                        .collect::<Result<Vec<Box<dyn SpanContents<'_>>>, MietteError>>()
+                        .map_err(|_| fmt::Error)?;
+                    let mut contexts = Vec::new();
+                    for (right, right_conts) in labels.iter().cloned().zip(contents.iter()) {
+                        if contexts.is_empty() {
+                            contexts.push((right, right_conts));
+                        } else {
+                            let (left, left_conts) = contexts.last().unwrap().clone();
+                            let left_end = left.offset() + left.len();
+                            let right_end = right.offset() + right.len();
+                            if left_conts.line() + left_conts.line_count() >= right_conts.line() {
+                                // The snippets will overlap, so we create one Big Chunky Boi
+                                let new_span = LabeledSpan::new(
+                                    left.label().map(String::from),
+                                    left.offset(),
+                                    if right_end >= left_end {
+                                        // Right end goes past left end
+                                        right_end - left.offset()
+                                    } else {
+                                        // right is contained inside left
+                                        left.len()
+                                    },
+                                );
+                                if source
+                                    .read_span(
+                                        new_span.inner(),
+                                        self.context_lines,
+                                        self.context_lines,
+                                    )
+                                    .is_ok()
+                                {
+                                    contexts.pop();
+                                    contexts.push((
+                                        new_span, // We'll throw this away later
+                                        left_conts,
+                                    ));
+                                } else {
+                                    contexts.push((right, right_conts));
+                                }
+                            } else {
+                                contexts.push((right, right_conts));
+                            }
+                        }
+                    }
+                    for (ctx, _) in contexts {
+                        self.render_context(f, source, &ctx, &labels[..])?;
+                    }
+                }
+            }
+        }
+        Ok(())
+    }
+
+    fn render_context<'a>(
+        &self,
+        f: &mut impl fmt::Write,
+        source: &'a dyn SourceCode,
+        context: &LabeledSpan,
+        labels: &[LabeledSpan],
+    ) -> fmt::Result {
+        let (contents, lines) = self.get_lines(source, context.inner())?;
+
+        // sorting is your friend
+        let labels = labels
+            .iter()
             .zip(self.theme.styles.highlights.iter().cloned().cycle())
-            .map(|((label, hl), st)| FancySpan::new(label, hl, st))
+            .map(|(label, st)| {
+                FancySpan::new(label.label().map(String::from), label.inner().clone(), st)
+            })
             .collect::<Vec<_>>();
 
         // The max number of gutter-lines that will be active at any given
@@ -217,7 +325,7 @@ impl GraphicalReportHandler {
         let mut max_gutter = 0usize;
         for line in &lines {
             let mut num_highlights = 0;
-            for hl in &highlights {
+            for hl in &labels {
                 if !line.span_line_only(hl) && line.span_applies(hl) {
                     num_highlights += 1;
                 }
@@ -239,24 +347,23 @@ impl GraphicalReportHandler {
             "{}{}{}",
             " ".repeat(linum_width + 2),
             self.theme.characters.ltop,
-            self.theme.characters.hbar.to_string().repeat(3),
+            self.theme.characters.hbar,
         )?;
-        if let Some(source_name) = snippet.source.name() {
-            let source_name = source_name.style(self.theme.styles.filename);
-            write!(
+
+        if let Some(source_name) = contents.name() {
+            let source_name = source_name.style(self.theme.styles.link);
+            writeln!(
                 f,
                 "[{}:{}:{}]",
                 source_name,
                 contents.line() + 1,
                 contents.column() + 1
             )?;
+        } else if lines.len() == 1 {
+            writeln!(f, "{}", self.theme.characters.hbar.to_string().repeat(3))?;
         } else {
-            write!(f, "[{}:{}]", contents.line() + 1, contents.column() + 1)?;
+            writeln!(f, "[{}:{}]", contents.line() + 1, contents.column() + 1)?;
         }
-        if let Some(msg) = &snippet.message {
-            write!(f, " {}:", msg)?;
-        }
-        writeln!(f)?;
 
         // Now it's time for the fun part--actually rendering everything!
         for line in &lines {
@@ -266,13 +373,13 @@ impl GraphicalReportHandler {
             // Then, we need to print the gutter, along with any fly-bys We
             // have separate gutters depending on whether we're on the actual
             // line, or on one of the "highlight lines" below it.
-            self.render_line_gutter(f, max_gutter, line, &highlights)?;
+            self.render_line_gutter(f, max_gutter, line, &labels)?;
 
             // And _now_ we can print out the line text itself!
             writeln!(f, "{}", line.text)?;
 
             // Next, we write all the highlights that apply to this particular line.
-            let (single_line, multi_line): (Vec<_>, Vec<_>) = highlights
+            let (single_line, multi_line): (Vec<_>, Vec<_>) = labels
                 .iter()
                 .filter(|hl| line.span_applies(hl))
                 .partition(|hl| line.span_line_only(hl));
@@ -280,14 +387,14 @@ impl GraphicalReportHandler {
                 // no line number!
                 self.write_no_linum(f, linum_width)?;
                 // gutter _again_
-                self.render_highlight_gutter(f, max_gutter, line, &highlights)?;
+                self.render_highlight_gutter(f, max_gutter, line, &labels)?;
                 self.render_single_line_highlights(
                     f,
                     line,
                     linum_width,
                     max_gutter,
                     &single_line,
-                    &highlights,
+                    &labels,
                 )?;
             }
             for hl in multi_line {
@@ -295,7 +402,7 @@ impl GraphicalReportHandler {
                     // no line number!
                     self.write_no_linum(f, linum_width)?;
                     // gutter _again_
-                    self.render_highlight_gutter(f, max_gutter, line, &highlights)?;
+                    self.render_highlight_gutter(f, max_gutter, line, &labels)?;
                     self.render_multi_line_end(f, hl)?;
                 }
             }
@@ -305,7 +412,7 @@ impl GraphicalReportHandler {
             "{}{}{}",
             " ".repeat(linum_width + 2),
             self.theme.characters.lbot,
-            self.theme.characters.hbar.to_string().repeat(3),
+            self.theme.characters.hbar.to_string().repeat(4),
         )?;
         Ok(())
     }
@@ -409,7 +516,7 @@ impl GraphicalReportHandler {
         write!(
             f,
             " {:width$} {} ",
-            linum,
+            linum.style(self.theme.styles.linum),
             self.theme.characters.vbar,
             width = width
         )?;
@@ -471,23 +578,39 @@ impl GraphicalReportHandler {
         }
         writeln!(f, "{}", underlines)?;
 
-        for hl in single_liners {
+        let vbar_offsets: Vec<_> = single_liners
+            .iter()
+            .map(|hl| {
+                (
+                    hl,
+                    (hl.offset() - line.offset) + (std::cmp::max(1, hl.len()) / 2),
+                )
+            })
+            .collect();
+        for hl in single_liners.iter().rev() {
             if let Some(label) = hl.label() {
                 self.write_no_linum(f, linum_width)?;
                 self.render_highlight_gutter(f, max_gutter, line, all_highlights)?;
-                let hl_len = std::cmp::max(1, hl.len());
-                let local_offset = hl.offset() - line.offset;
-                let vbar_offset = local_offset + (hl_len / 2);
-                let num_right = local_offset + hl_len - vbar_offset - 1;
-                let lines = format!(
-                    "{:width$}{}{} {}",
-                    "",
-                    chars.lbot,
-                    chars.hbar.to_string().repeat(num_right + 1),
-                    label,
-                    width = vbar_offset
-                );
-                writeln!(f, "{}", lines.style(hl.style))?;
+                let mut curr_offset = 1usize;
+                for (offset_hl, vbar_offset) in &vbar_offsets {
+                    while curr_offset < *vbar_offset + 1 {
+                        write!(f, " ")?;
+                        curr_offset += 1;
+                    }
+                    if *offset_hl != hl {
+                        write!(f, "{}", chars.vbar.to_string().style(offset_hl.style))?;
+                        curr_offset += 1;
+                    } else {
+                        let lines = format!(
+                            "{}{} {}",
+                            chars.lbot,
+                            chars.hbar.to_string().repeat(2),
+                            label,
+                        );
+                        writeln!(f, "{}", lines.style(hl.style))?;
+                        break;
+                    }
+                }
             }
         }
         Ok(())
@@ -505,16 +628,16 @@ impl GraphicalReportHandler {
 
     fn get_lines<'a>(
         &'a self,
-        snippet: &'a DiagnosticSnippet<'_>,
-    ) -> Result<(Box<dyn SpanContents + 'a>, Vec<Line>), fmt::Error> {
-        let context_data = snippet
-            .source
-            .read_span(&snippet.context)
+        source: &'a dyn SourceCode,
+        context_span: &'a SourceSpan,
+    ) -> Result<(Box<dyn SpanContents<'a> + 'a>, Vec<Line>), fmt::Error> {
+        let context_data = source
+            .read_span(context_span, self.context_lines, self.context_lines)
             .map_err(|_| fmt::Error)?;
         let context = std::str::from_utf8(context_data.data()).expect("Bad utf8 detected");
         let mut line = context_data.line();
         let mut column = context_data.column();
-        let mut offset = snippet.context.offset();
+        let mut offset = context_data.span().offset();
         let mut line_offset = offset;
         let mut iter = context.chars().peekable();
         let mut line_str = String::new();
@@ -578,6 +701,7 @@ impl ReportHandler for GraphicalReportHandler {
 Support types
 */
 
+#[derive(Debug)]
 struct Line {
     line_number: usize,
     offset: usize,
@@ -592,11 +716,11 @@ impl Line {
 
     fn span_applies(&self, span: &FancySpan) -> bool {
         // Span starts in this line
-        (span.offset() >= self.offset && span.offset() < self.offset +self.length)
+        (span.offset() >= self.offset && span.offset() < self.offset + self.length)
         // Span passes through this line
         || (span.offset() < self.offset && span.offset() + span.len() > self.offset + self.length) //todo
         // Span ends on this line
-        || (span.offset() + span.len() >= self.offset && span.offset() + span.len() <= self.offset + self.length)
+        || (span.offset() + span.len() > self.offset && span.offset() + span.len() <= self.offset + self.length)
     }
 
     // A "flyby" is a multi-line span that technically covers this line, but
@@ -623,12 +747,19 @@ impl Line {
     }
 }
 
+#[derive(Debug, Clone)]
 struct FancySpan {
     label: Option<String>,
     span: SourceSpan,
     style: Style,
 }
 
+impl PartialEq for FancySpan {
+    fn eq(&self, other: &Self) -> bool {
+        self.label == other.label && self.span == other.span
+    }
+}
+
 impl FancySpan {
     fn new(label: Option<String>, span: SourceSpan, style: Style) -> Self {
         FancySpan { label, span, style }
